#!/bin/python3

__author__="Raphael Leu"
__version__="@VERSION@"

from .data_structures import *
from .indicators import *
from .io_helpers import *
from .mesh import *
from .refinement import *
from .solvers import *



import numpy as np
from p1afempy.mesh import get_directional_vectors, get_area
from scipy.sparse import coo_matrix, csc_matrix
from scipy.sparse.linalg import spsolve
from p1afempy.data_structures import \
    CoordinatesType, ElementsType, BoundaryConditionType, BoundaryType


def get_stiffness_matrix(coordinates: CoordinatesType,
                         elements: ElementsType) -> coo_matrix:
    """
    returns the stiffness matrix for the P1 FEM
    with Legendre basis

    parameters
    ----------
    coordinates: CoordinatesType
    elements: ElementsType

    returns
    -------
    scipy.sparse.coo_matrix: the sparse stiffness matrix
    """
    # vector of element areas 4*|T|
    area4 = 4. * get_area(coordinates=coordinates,
                          elements=elements)

    indices_i = (elements[:, [0, 1, 2, 0, 1, 2, 0, 1, 2]].T).flatten(
        order='F')
    indices_j = (elements[:, [0, 0, 0, 1, 1, 1, 2, 2, 2]].T).flatten(
        order='F')

    d21, d31 = get_directional_vectors(coordinates=coordinates,
                                       elements=elements)
    a = (np.sum(d21*d31, axis=1)/area4)
    b = (np.sum(d31*d31, axis=1)/area4)
    c = (np.sum(d21*d21, axis=1)/area4)

    A = np.vstack([-2.*a+b+c, a-b, a-c, a-b, b, -a, a-c, -a, c])
    return coo_matrix((A.flatten(order='F'),
                       (indices_i, indices_j)))


def get_mass_matrix(coordinates: CoordinatesType,
                    elements: ElementsType) -> coo_matrix:
    """
    returns the mass matrix of the mesh provided
    for the P1 FEM with Legendre basis
    """
    I, J, D = get_mass_matrix_elements(coordinates=coordinates,
                                       elements=elements)
    return coo_matrix((D, (I, J)))


def get_mass_matrix_elements(
        coordinates: CoordinatesType,
        elements: ElementsType) -> tuple[np.ndarray, np.ndarray, np.ndarray]:
    """
    returns the mass matrix of the mesh provided
    for the P1 FEM with Legendre basis

    returns
    -------
    indices_i: np.ndarray
    indices_j: np.ndarray
    D: np.ndarray
        D[m] represents a mass matrix contribution
        belonging to its (indices_i[m], indices_j[m]) coordinate
    """
    indices_i = (elements[:, [0, 1, 2, 0, 1, 2, 0, 1, 2]].T).flatten(
        order='F')
    indices_j = (elements[:, [0, 0, 0, 1, 1, 1, 2, 2, 2]].T).flatten(
        order='F')

    area = get_area(coordinates=coordinates,
                    elements=elements)
    D = np.vstack(
        [area/6., area/12., area/12.,
         area/12., area/6., area/12.,
         area/12., area/12., area/6.]).flatten(order='F')
    return indices_i, indices_j, D


def get_right_hand_side(coordinates: CoordinatesType,
                        elements: ElementsType,
                        f: BoundaryConditionType):
    """
    returns the load vector for the P1 FEM with Legendre basis

    parameters
    ----------
    coordinates: CoordinatesType
    elements: ElementsType
    f: BoundaryConditionType
        the function for which to evaluate the load vector

    returns
    -------
    b: np.ndarray
        the P1 FEM load vector of f on the mesh at hand

    notes
    -----
    the load vector F_i := int f(x)phi_i(x) dx
    is approximated as sum_T |T| * f(sT) * phi_i(sT),
    where sT denotes the center of mass of triangle T
    """
    # vector of element areas 4*|T|
    area4 = 4. * get_area(coordinates=coordinates,
                          elements=elements)

    # assembly of right-hand side
    d21, d31 = get_directional_vectors(coordinates=coordinates,
                                       elements=elements)
    fsT = f((coordinates[elements[:, 0], :]+(d21+d31) / 3))
    b = np.bincount(
        elements.flatten(order='F'),
        weights=np.tile(area4*fsT/12., (3, 1)).flatten(),
        minlength=coordinates.shape[0])
    return b


def apply_neumann(neumann_bc: BoundaryType,
                  coordinates: CoordinatesType,
                  g: BoundaryConditionType,
                  b: np.ndarray):
    """
    applies neuman boundary conditions to b and returns new b
    """
    # TODO channge b in place, do not return it or
    # at least check if this version generates computational overhead
    cn1 = coordinates[neumann_bc[:, 0], :]
    cn2 = coordinates[neumann_bc[:, 1], :]
    gmE = g((cn1+cn2)/2)
    b = b + np.bincount(
        neumann_bc.flatten(order='F'),
        weights=np.tile(
            np.sqrt(np.sum(np.square(cn2-cn1), axis=1))*gmE/2.,
            (2, 1)).flatten(), minlength=b.size)
    return b


def solve_laplace(coordinates: CoordinatesType,
                  elements: ElementsType,
                  dirichlet: BoundaryType,
                  neumann: BoundaryType,
                  f: BoundaryConditionType,
                  g: BoundaryConditionType,
                  uD: BoundaryConditionType
                  ) -> tuple[np.ndarray, float]:
    """
    solves the laplace equation, i.e.

    -Delta u = f, on Omega
    u = uD, on Gamma_D
    du/dn = g, on Gamma_N

    on the provided mesh using P1 FEM with Legendre basis

    parameters
    ----------
    coordinates: CoordinatesType
    elements: ElementsType
    dirichlet: BoundaryType
        the dirichlet boundary of the problem
    neumann: BoundaryType
        the neumann boundary of the problem
    f: BoundaryConditionType
        the right-hand-side function (volume force) of the problem
    g: BoundaryConditionType
        the neumann boundary function, i.e.
        u(x) = g(x) on Gamma_N
    uD: BoundaryConditionType
        the dirichlet boundary function, i.e.
        du/dn(x) = uD(x) on Gamma_D

    returns
    -------
    x: np.ndarray
        Nx2 array representing the solution of the
        given laplace problem on the mesh provided
    energy: float
        the energy (A-norm: x.T*A*x) of the solution found

    notes
    -----
    the functions f, g, and uD are all expected to be callable like
    f(coordinates), where coordinates is an (n_coordinates x 2) array
    """
    n_coordinates = coordinates.shape[0]
    x = np.zeros(n_coordinates)

    A = get_stiffness_matrix(coordinates=coordinates,
                             elements=elements)

    # prescribe values at dirichlet nodes
    unique_dirichlet = np.unique(dirichlet)
    x[unique_dirichlet] = uD((coordinates[unique_dirichlet, :]))

    b = get_right_hand_side(coordinates=coordinates,
                            elements=elements, f=f) - A.dot(x)
    if neumann.size > 0:
        b = apply_neumann(neumann_bc=neumann,
                          coordinates=coordinates,
                          g=g, b=b)

    # computation of P1-FEM approximation
    freenodes = np.setdiff1d(
        np.arange(0, n_coordinates), unique_dirichlet, assume_unique=True)
    A = csc_matrix(A)
    x[freenodes] = spsolve(A[freenodes, :][:, freenodes],
                           b[freenodes],
                           use_umfpack=True)
    # compute energy || grad(uh) ||^2 of discrete solution
    energy = x.dot(A.dot(x))

    return x, energy



import numpy as np
from pathlib import Path


def read_boundary_condition(path_to_boundary: Path,
                            shift_indices: bool = False) -> np.ndarray:
    """
    Reads boundary condition data from a file.

    Parameters
    ----------
    path_to_boundary: pathlib.Path
        Path to the file containing boundary condition data.
    shift_indices: bool (default=False)
        If true, shifts the elements' indices according to i':=i-1.
        This can come in handy if, e.g., one wants to read data
        that is compatible with Matlab/Fortran/Julia indexing
        (starting at 1, instead of 0).

    Returns
    -------
    boundary_indices: np.ndarray:
        boundary conditions as np.nndarray, i.e.
        boundary_indices[k] gives the indices (i_k, j_k) of the
        vertices belonging to the k-th edge of the boundary condition at hand

    Example
    -------
    >>> file_path = Path("path/to/boundary_data.dat")
    >>> boundary_condition = read_boundary_condition(file_path)
    """
    boundary_indices = np.loadtxt(path_to_boundary).astype(np.uint32)
    if shift_indices:
        boundary_indices = boundary_indices - 1
    return boundary_indices


def read_coordinates(path_to_coordinates: Path) -> np.ndarray:
    """
    reads and returns vertices from file

    Parameters
    ----------
    path_to_coordinates : pathlib.Path
        Path to the file containing mesh coordinates data.
    path_to_elements : pathlib.Path

    Returns
    -------
    coordinates: CoordinatesType
        coordinates of the mesh's vertices

    Example
    -------
    >>> coordinates_path = Path("path/to/coordinates.txt")
    >>> coordinates = read_coordinates(coordinates_path)
    """
    coordinates = np.loadtxt(path_to_coordinates)
    return coordinates


def read_elements(path_to_elements: Path,
                  shift_indices: bool = False) -> np.ndarray:
    """
    reads and returns elements from file.

    Parameters
    ----------
    path_to_elements : pathlib.Path
        Path to the file containing mesh elements data.
    shift_indices: bool (default=False)
        If true, shifts the elements' indices according to i':=i-1.
        This can come in handy if, e.g., one wants to read data
        that is compatible with Matlab/Fortran/Julia indexing
        (starting at 1, instead of 0).

    Returns
    -------
    elements: ElementsType
        elements of the mesh

    Example
    -------
    >>> elements_path = Path("path/to/elements_from_matlab_code.txt")
    >>> elements = read_elements(elements_path, shift_indices=True)
    """
    elements = np.loadtxt(path_to_elements).astype(np.uint32)
    if shift_indices:
        elements = elements - 1
    return elements


def read_mesh(path_to_coordinates: Path,
              path_to_elements: Path,
              shift_indices: bool = False) -> tuple[np.ndarray, np.ndarray]:
    """
    Reads vertices and elements from files and
    returns the corresponding Mesh.

    Parameters
    ----------
    path_to_coordinates : pathlib.Path
        Path to the file containing mesh coordinates data.
    path_to_elements : pathlib.Path
        Path to the file containing mesh elements data.
    shift_indices: bool (default=False)
        If true, shifts the elements' indices according to i':=i-1.
        This can come in handy if, e.g., one wants to read data
        that is compatible with Matlab/Fortran/Julia indexing
        (starting at 1, instead of 0).

    Returns
    -------
    coordinates: CoordinatesType
        coordinates of the mesh's vertices
    elements: ElementsType
        elements of the mesh

    Example
    -------
    >>> coordinates_path = Path("path/to/coordinates.txt")
    >>> elements_path = Path("path/to/elements.txt")
    >>> coordinates, elements = read_mesh(coordinates_path, elements_path)
    """
    coordinates = read_coordinates(path_to_coordinates=path_to_coordinates)
    elements = read_elements(path_to_elements=path_to_elements,
                             shift_indices=shift_indices)
    return coordinates, elements



import numpy as np
from typing import Callable

"""
This files serves as a lookup dictionary on the data structures
used throughout this package. Note that the data structres used
here coincide with the data structures defined in [1].

References
----------
- [1] S. Funken, D. Praetorius, and P. Wissgott.
      Efficient Implementation of Adaptive P1-FEM in Matlab,
      (http://dx.doi.org/10.2478/cmam-2011-0026).
      Computational Methods in Applied Mathematics,
      Vol. 11 (2011), No. 4, pp. 460–490.
"""

ElementsType = np.ndarray
"""
The triangulation T is represented by the Mx3 integer array elements.
The l-th triangle T_l = conv{zi, zj , zk} ∈ T with vertices zi, zj , zk ∈ N
is stored as elements[l, :] = [i, j, k],
where the nodes are given in counterclockwise order,
i.e. the parametrization of the boundary is mathematically positive.
"""

CoordinatesType = np.ndarray
"""
The set of all nodes is represented by the Nx2 array coordinates.
The k-th row of coordinates stores the coordinates of the k-th node, i.e.
z_l = (x_l, y_l) ∈ R2 as coordinates[k, :] = [x_k, y_k].
"""

BoundaryType = np.ndarray
"""
The Dirichlet boundary is split into K affine boundary pieces,
which are edges of triangles T ∈ T.
It is represented by a Kx2 integer array dirichlet.
The l-th edge E_l = conv{zi, zj} on the Dirichlet boundary is stored
in the form dirichlet[l, :] = [i, j].
"""

BoundaryConditionType = Callable[[CoordinatesType], np.ndarray]
"""
Any function specifying a boundary condition is expected to
be implemented such that it can be called on the Nx2 array coordinates
and returns an Nx1 array, where the function was applied row-wise.
"""



import numpy as np
from p1afempy.mesh import provide_geometric_data
from p1afempy.data_structures import \
    CoordinatesType, ElementsType, BoundaryType


def generate_new_nodes(edge2newNode: np.ndarray,
                       edge2nodes: np.ndarray,
                       coordinates: CoordinatesType,
                       to_embed: np.ndarray = np.array([])) -> tuple[
                           np.ndarray,
                           np.ndarray]:
    n_new_nodes = np.count_nonzero(edge2newNode)  # number of new nodes
    # assigning indices to new nodes
    edge2newNode[edge2newNode != 0] = np.arange(
        coordinates.shape[0],
        coordinates.shape[0] + n_new_nodes)
    edges_with_new_nodes_indices = np.nonzero(edge2newNode)[0]
    new_node_coordinates = (
        coordinates[edge2nodes[edges_with_new_nodes_indices, 0], :] +
        coordinates[edge2nodes[edges_with_new_nodes_indices, 1], :]) / 2.

    # interpolate values, if given
    if to_embed.size:
        new_embedded_values = (
            to_embed[edge2nodes[edges_with_new_nodes_indices, 0]] +
            to_embed[edge2nodes[edges_with_new_nodes_indices, 1]]) / 2.
        to_embed = np.hstack([to_embed, new_embedded_values])

    return np.vstack([coordinates, new_node_coordinates]), to_embed


def refineRG(coordinates: CoordinatesType,
             elements: ElementsType,
             marked_element: int,
             boundaries: list[BoundaryType],
             to_embed: np.ndarray = np.array([])) -> tuple[CoordinatesType,
                                                           ElementsType,
                                                           list[BoundaryType]]:
    """
    refines the mesh according to
    red-green refinement of one single element

    parameters
    ----------
    coordinates: CoordinatesType
    elements: ElementsType
    marked_element: int
        the index of the element to be red refined
    boundaries: list[BoundaryType]
    to_embed: np.ndarray = np.array([])
        vector of values on coordinates to be interpolated
        (canonically embedded) onto the refined mesh

    returns
    -------
    new_coordinates: CoordinatesType
        coordinates of the refined mesh
    new_elements: ElementsType
        elements of the refined mesh
    new_boundaries: list[BoundaryType]
        boundaries of the refines mesh
    embedded_values: np.ndarray
        to_embed interpolated onto the refined mesh

    notes
    -----
    this method does the following
    - given an element k=marked_element marked for
      refinement
    - red refine element k
    - green refine all neighbouring elements (at most three),
      meaning that the new vertex on the edge shared by k
      and its neighbour k' gets connected to the vertex
      in k' opposite to the shared edge
    - if any new vertex lies on a boundary,
      refine the boundary accordingly
    """
    nE = elements.shape[0]

    # Obtain geometric information on edges
    element2edges, edge2nodes, boundary2edges = \
        provide_geometric_data(elements=elements, boundaries=boundaries)

    # Mark edges for refinement
    edge2newNode = np.zeros(edge2nodes.shape[0], dtype=int)
    edge2newNode[element2edges[marked_element].flatten()] = 1

    new_coordinates, embedded_values = generate_new_nodes(
        edge2newNode=edge2newNode,
        edge2nodes=edge2nodes,
        coordinates=coordinates,
        to_embed=to_embed)

    # refine boundary conditions
    new_boundaries = []
    for k, boundary in enumerate(boundaries):
        if boundary.size:
            new_nodes_on_boundary = edge2newNode[boundary2edges[k]]
            marked_edges = np.nonzero(new_nodes_on_boundary)[0]
            if marked_edges.size:
                boundary = np.vstack(
                    [boundary[np.logical_not(new_nodes_on_boundary), :],
                     np.column_stack([boundary[marked_edges, 0],
                                      new_nodes_on_boundary[marked_edges]]),
                     np.column_stack([new_nodes_on_boundary[marked_edges],
                                      boundary[marked_edges, 1]])])
        new_boundaries.append(boundary)

    # Provide new nodes for refinement of elements
    newNodes = edge2newNode[element2edges]

    # Determine type of refinement for each element
    marked_edges = newNodes != 0

    first_marked = marked_edges[:, 0]
    second_marked = marked_edges[:, 1]
    third_marked = marked_edges[:, 2]

    not_first = np.logical_not(first_marked)
    not_second = np.logical_not(second_marked)
    not_third = np.logical_not(third_marked)

    none = not_first & not_second & not_third
    green1 = first_marked & not_second & not_third
    green2 = not_first & second_marked & not_third
    green3 = not_first & not_second & third_marked
    red = first_marked & second_marked & third_marked

    # generate element numbering for refined mesh
    idx = np.ones(nE, dtype=int)
    idx[green1] = 2  # green(1): green refinement of 1st edge
    idx[green2] = 2  # green(2): green refinement of 2nd edge
    idx[green3] = 2  # green(3): green refinement of 3rd edge
    idx[red] = 4  # red: red refinement
    idx = np.hstack([0, np.cumsum(idx)])

    # generate new elements
    # ---------------------
    new_elements = np.zeros((idx[-1], 3), dtype=int)
    idx = idx[:-1]

    # no refinement
    new_elements[idx[none], :] = elements[none, :]

    # green refinement (1)
    new_elements[np.hstack([idx[green1], 1+idx[green1]]), :] \
        = np.vstack(
            [np.column_stack([
                elements[green1, 0],
                newNodes[green1, 0],
                elements[green1, 2]]),
             np.column_stack([
                 newNodes[green1, 0],
                 elements[green1, 1],
                 elements[green1, 2]])])

    # green refinement (2)
    new_elements[np.hstack([idx[green2], 1+idx[green2]]), :] \
        = np.vstack(
            [np.column_stack([
                elements[green2, 1],
                newNodes[green2, 1],
                elements[green2, 0]]),
             np.column_stack([
                 newNodes[green2, 1],
                 elements[green2, 2],
                 elements[green2, 0]])])

    # green refinement (3)
    new_elements[np.hstack([idx[green3], 1+idx[green3]]), :] \
        = np.vstack(
            [np.column_stack([
                elements[green3, 2],
                newNodes[green3, 2],
                elements[green3, 1]]),
             np.column_stack([
                 newNodes[green3, 2],
                 elements[green3, 0],
                 elements[green3, 1]])])

    # red refinement
    new_elements[np.hstack([idx[red], 1+idx[red],
                            2+idx[red], 3+idx[red]]), :] \
        = np.vstack([
            np.column_stack([elements[red, 0],
                             newNodes[red, 0],
                             newNodes[red, 2]]),
            np.column_stack([newNodes[red, 0],
                             elements[red, 1],
                             newNodes[red, 1]]),
            np.column_stack([newNodes[red, 2],
                             newNodes[red, 1],
                             elements[red, 2]]),
            np.column_stack([newNodes[red, 0],
                             newNodes[red, 1],
                             newNodes[red, 2]])])

    return new_coordinates, new_elements, new_boundaries, embedded_values


# TODO refactor s.t. boundary_conditions is optional
def refineRGB(coordinates: CoordinatesType,
              elements: ElementsType,
              marked_elements: np.ndarray,
              boundary_conditions: list[BoundaryType],
              to_embed: np.ndarray = np.array([])
              ) -> tuple[CoordinatesType,
                         ElementsType,
                         list[BoundaryType]]:
    """
    Refines the mesh and boundary conditions based on the
    red-green-blue (RGB) refinement strategy.

    Parameters
    ----------
    coordinates: CoordinatesType
    elements: ElementsType
    marked_elements: np.ndarray
        Indices of the elements to be refined.
    boundary_conditions: list[BoundaryType]
        List of boundary conditions to update.
    to_embed: np.ndarray = np.array([])
        vector of values on coordinates to be interpolated
        (canonically embedded) onto the refined mesh

    Returns
    -------
    new_coordinates: CoordinatesType
        the coordinates of the refined mesh
    new_elements: ElementsType
        the elements of the refined mesh
    new_boundaries: list[BoundaryType]
        The refined boundary conditions
    embedded_values: np.ndarray
        to_embed interpolated onto the refined mesh

    Example
    -------
    >>> coordinates, elements = ...  # Initialize a mesh
    >>> marked_elements = np.array([0, 2, 3, 4])
    >>> boundary_conditions = [BC1, BC2, BC3]  # BCs as `np.ndarray`s
    >>> new_coordinates, new_elements, new_boundary_conditions, _ = refineRGB(
            coordinates, elements,
            marked_elements, boundary_conditions)
    """
    new_coordinates, new_elements, new_boundaries, embedded_values = refineNVB(
        coordinates,
        elements,
        marked_elements,
        boundary_conditions=boundary_conditions,
        sort_for_longest_egde=True,
        to_embed=to_embed)
    return new_coordinates, new_elements, new_boundaries, embedded_values


# TODO refactor s.t. sort_for_longest_egde vanishes, this is an ugly solution
# TODO refactor s.t. boundary_conditions is optional
def refineNVB(coordinates: CoordinatesType,
              elements: ElementsType,
              marked_elements: np.ndarray,
              boundary_conditions: list[BoundaryType],
              to_embed: np.ndarray = np.array([]),
              sort_for_longest_egde: bool = False
              ) -> tuple[CoordinatesType,
                         ElementsType,
                         list[BoundaryType],
                         np.ndarray]:
    """
    refines the mesh based on newest vertex bisection (NVB)

    Parameters
    ----------
    coordinates: CoordinatesType
    elements: ElementsType
    marked_elements: np.ndarray
        indices of the elements to be refined
    boundary_conditions: list[BoundaryType]
        list of boundaries to be refined
    to_embed: np.ndarray = np.array([])
        vector of values on coordinates to be interpolated
        (canonically embedded) onto the refined mesh

    Returns
    -------
    new_coordinates: CoordinatesType
        the coordinates of the refined mesh
    new_elements: ElementsType
        the elements of the refined mesh
    new_boundaries: list[BoundaryType]
        the refined boundary conditions
    embedded_values: np.ndarray
        to_embed interpolated onto the refined mesh

    Example
    -------
    >>> coordinates, elements = Mesh(...)  # Initialize a mesh
    >>> marked_elements = np.array([0, 2, 3, 4])
    >>> boundary_conditions = [BC1, BC2, BC3]  # BC's as np.ndarray
    >>> new_coordinates, new_elements, new_boundary_conditions, _ = refineNVB(
            mesh.coordinates, mesh.elements,
            marked_elements, boundary_conditions)
    """
    n_elements = elements.shape[0]

    if sort_for_longest_egde:
        # Sort elements such that first edge is longest
        dx = (coordinates[elements[:, [1, 2, 0]], 0]
              - coordinates[elements, 0]).flatten(order='F')
        dy = (coordinates[elements[:, [1, 2, 0]], 1]
              - coordinates[elements, 1]).flatten(order='F')
        idxMax = np.argmax(
            (np.square(dx)+np.square(dy)).reshape((n_elements, 3), order='F'),
            axis=1)
        idx = idxMax == 1
        elements[idx, :] = elements[idx][:, [1, 2, 0]]
        idx = idxMax == 2
        elements[idx, :] = elements[idx][:, [2, 0, 1]]

    # obtain geometric information on edges
    element2edges, edge_to_nodes, boundaries_to_edges = provide_geometric_data(
        elements=elements,
        boundaries=boundary_conditions)

    # mark all edges of marked elements for refinement
    edge2newNode = np.zeros(edge_to_nodes.shape[0], dtype=int)
    edge2newNode[element2edges[marked_elements].flatten()] = 1

    # closure of edge marking, i.e.
    # if any edge in T is marked, make sure that the reference
    # edge in T is marked, as well
    swap = np.array([1])
    while swap.size > 0:
        element2marked_edges = edge2newNode[element2edges]
        swap = np.nonzero(
            np.logical_and(
                # elements, whose reference edge is not marked
                np.logical_not(element2marked_edges[:, 0]),
                # elements, having any non-reference edge marked
                np.logical_or(element2marked_edges[:, 1],
                              element2marked_edges[:, 2])))[0]
        edge2newNode[element2edges[swap, 0]] = 1

    new_coordinates, embedded_values = generate_new_nodes(
        edge2newNode=edge2newNode,
        edge2nodes=edge_to_nodes,
        coordinates=coordinates,
        to_embed=to_embed)

    # refine boundary conditions
    new_boundaries = []
    for k, boundary in enumerate(boundary_conditions):
        if boundary.size:
            new_nodes_on_boundary = edge2newNode[boundaries_to_edges[k]]
            marked_edges = np.nonzero(new_nodes_on_boundary)[0]
            if marked_edges.size:
                boundary = np.vstack(
                    [boundary[np.logical_not(new_nodes_on_boundary), :],
                     np.column_stack([boundary[marked_edges, 0],
                                      new_nodes_on_boundary[marked_edges]]),
                     np.column_stack([new_nodes_on_boundary[marked_edges],
                                      boundary[marked_edges, 1]])])
        new_boundaries.append(boundary)

    # provide new nodes for refinement of elements
    new_nodes = edge2newNode[element2edges]

    # Determine type of refinement for each element
    marked_edges = new_nodes != 0

    ref_marked = marked_edges[:, 0]
    first_marked = marked_edges[:, 1]
    second_marked = marked_edges[:, 2]

    none = np.logical_not(ref_marked)
    not_first = np.logical_not(first_marked)
    not_second = np.logical_not(second_marked)
    bisec1 = ref_marked & not_first & not_second
    bisec12 = ref_marked & first_marked & not_second
    bisec13 = ref_marked & not_first & second_marked
    bisec123 = ref_marked & first_marked & second_marked

    # generate element numbering for refined mesh
    idx = np.ones(n_elements, dtype=int)
    idx[bisec1] = 2  # bisec(1): newest vertex bisection of 1st edge
    idx[bisec12] = 3  # bisec(2): newest vertex bisection of 1st and 2nd edge
    idx[bisec13] = 3  # bisec(2): newest vertex bisection of 1st and 3rd edge
    idx[bisec123] = 4  # bisec(3): newest vertex bisection of all edges
    idx = np.hstack([0, np.cumsum(idx)])  # TODO maybe bug source

    # generate new elements
    new_elements = np.zeros((idx[-1], 3), dtype=int)
    idx = idx[:-1]
    new_elements[idx[none], :] = elements[none, :]
    new_elements[np.hstack([idx[bisec1],
                           1+idx[bisec1]]), :] \
        = np.vstack(
            [np.column_stack([
                elements[bisec1, 2],
                elements[bisec1, 0],
                new_nodes[bisec1, 0]]),
             np.column_stack([
                 elements[bisec1, 1],
                 elements[bisec1, 2],
                 new_nodes[bisec1, 0]])])
    new_elements[np.hstack([idx[bisec12],
                           1+idx[bisec12],
                           2+idx[bisec12]]), :] \
        = np.vstack(
            [np.column_stack([elements[bisec12, 2],
                              elements[bisec12, 0],
                              new_nodes[bisec12, 0]]),
             np.column_stack([new_nodes[bisec12, 0],
                              elements[bisec12, 1],
                              new_nodes[bisec12, 1]]),
             np.column_stack([elements[bisec12, 2],
                              new_nodes[bisec12, 0],
                              new_nodes[bisec12, 1]])])
    new_elements[np.hstack([idx[bisec13],
                           1+idx[bisec13],
                           2+idx[bisec13]]), :] \
        = np.vstack(
            [np.column_stack([new_nodes[bisec13, 0],
                              elements[bisec13, 2],
                              new_nodes[bisec13, 2]]),
             np.column_stack([elements[bisec13, 0],
                              new_nodes[bisec13, 0],
                              new_nodes[bisec13, 2]]),
             np.column_stack([elements[bisec13, 1],
                              elements[bisec13, 2],
                              new_nodes[bisec13, 0]])])
    if sort_for_longest_egde:
        new_elements[np.hstack([idx[bisec123],
                               1+idx[bisec123],
                               2+idx[bisec123],
                               3+idx[bisec123]]), :] \
            = np.vstack([
                np.column_stack([elements[bisec123, 0],
                                 new_nodes[bisec123, 0],
                                 new_nodes[bisec123, 2]]),
                np.column_stack([new_nodes[bisec123, 0],
                                 elements[bisec123, 1],
                                 new_nodes[bisec123, 1]]),
                np.column_stack([new_nodes[bisec123, 2],
                                 new_nodes[bisec123, 1],
                                 elements[bisec123, 2]]),
                np.column_stack([new_nodes[bisec123, 1],
                                 new_nodes[bisec123, 2],
                                 new_nodes[bisec123, 0]])])
    else:
        new_elements[np.hstack([idx[bisec123],
                               1+idx[bisec123],
                               2+idx[bisec123],
                               3+idx[bisec123]]), :] \
            = np.vstack([
                np.column_stack([new_nodes[bisec123, 0],
                                elements[bisec123, 2],
                                new_nodes[bisec123, 2]]),
                np.column_stack([elements[bisec123, 0],
                                new_nodes[bisec123, 0],
                                new_nodes[bisec123, 2]]),
                np.column_stack([new_nodes[bisec123, 0],
                                elements[bisec123, 1],
                                new_nodes[bisec123, 1]]),
                np.column_stack([elements[bisec123, 2],
                                new_nodes[bisec123, 0],
                                new_nodes[bisec123, 1]])])

    return new_coordinates, new_elements, new_boundaries, embedded_values



import numpy as np
from p1afempy import data_structures
from scipy.sparse import coo_matrix, find
from matplotlib import pyplot as plt


def get_area(coordinates: data_structures.CoordinatesType,
             elements: data_structures.ElementsType) -> np.ndarray:
    """
    calculates and returns the area of each element as numpy array

    parameters
    ----------
    coordinates: data_structures.CoordinatesType
    elements: data_structures.ElementsType

    returns
    -------
    np.ndarray:
        area of each element as Mx1 array,
        where M represents the number ot elements.
    """
    d21, d31 = get_directional_vectors(coordinates=coordinates,
                                       elements=elements)
    return 0.5 * (d21[:, 0]*d31[:, 1] - d21[:, 1] * d31[:, 0])


def get_directional_vectors(coordinates: data_structures.CoordinatesType,
                            elements: data_structures.ElementsType
                            ) -> tuple[np.ndarray, np.ndarray]:
    """
    returns the vectors pointing from vertex[0] to vertex[1]
    and vertex[2], respectively, for each element

    Returns
    -------
    d21: np.ndarray
        Mx2 array, where d21[k, :] points from vertex 0 to vertex 1
        in the k-th element and M represents the number of elements
    d31: np.ndarray
        Mx2 array, where d31[k, :] points from vertex 0 to vertex 2
        in the k-th element and M represents the number of elements
    """
    c1 = coordinates[elements[:, 0], :]
    d21 = coordinates[elements[:, 1], :] - c1
    d31 = coordinates[elements[:, 2], :] - c1

    return d21, d31


def show_mesh(coordinates: data_structures.CoordinatesType,
              elements: data_structures.ElementsType) -> None:
    """displays the mesh at hand"""
    for element in elements:
        r0, r1, r2 = coordinates[element, :]
        plt.plot(
            [r0[0], r1[0], r2[0], r0[0]],
            [r0[1], r1[1], r2[1], r0[1]],
            'black', linewidth=0.5)
    plt.show()


def provide_geometric_data(elements: data_structures.ElementsType,
                           boundaries: list[data_structures.BoundaryType]
                           ) -> tuple[np.ndarray,
                                      np.ndarray,
                                      list[np.ndarray]]:
    """
    provides geometric data about the mesh (elements and boundaries) at hand

    Parameters
    ----------
    elements: data_structures.ElementsType
    boundaries: list[data_structures.BoundaryType

    Returns
    -------
    element_to_edges: np.ndarray
        element_to_edges[k] holds the edges' indices of
        the k-th element (counter-clockwise)
    edge_to_nodes: np.ndarray
        edge_to_nodes[k] holds the nodes' indices (i, j)
        of the k-th edge s.t. i < j
    boundaries_to_edges: list[np.ndarray]
        boundaries_to_edges[k] holds the mapping
        s.t. boundaries_to_edges[k][n] gives the indices
        (i, j) of the n-th edge of the k-th boundary.
    """
    n_elements = elements.shape[0]
    n_boundaries = len(boundaries)

    # Extracting all directed edges E_l:=(I[l], J[l])
    # (interior edges appear twice)
    element_indices_i = elements.flatten(order='F')
    element_indices_j = elements[:, [1, 2, 0]].flatten(order='F')

    # Symmetrize I and J (so far boundary edges appear only once)
    pointer = np.concatenate(([0, 3*n_elements-1],
                              np.zeros(n_boundaries, dtype=int)), dtype=int)
    for k, boundary in enumerate(boundaries):
        if boundary.size:
            element_indices_i = np.concatenate(
                (element_indices_i, boundary[:, 1]), dtype=int)
            element_indices_j = np.concatenate(
                (element_indices_j, boundary[:, 0]), dtype=int)
        pointer[k+2] = pointer[k+1] + boundary.shape[0]

    # Fixing an edge number for all edges, where i<j
    idx_IJ = np.where(element_indices_i < element_indices_j)[0]
    n_unique_edges = idx_IJ.size
    edge_number = np.zeros(element_indices_i.size, dtype=int)
    edge_number[idx_IJ] = np.arange(n_unique_edges)

    # Ensuring the same numbering for all edges, where j<i
    idx_JI = np.where(element_indices_j < element_indices_i)[0]
    number_to_edges = coo_matrix(
        (np.arange(n_unique_edges) + 1, (element_indices_i[idx_IJ],
                                         element_indices_j[idx_IJ])))
    # NOTE In Matlab, the returned order is different
    _, _, numbering_IJ = find(number_to_edges)
    # NOTE In Matlab, the returned order is different
    _, _, idx_JI2IJ = find(
        coo_matrix((idx_JI + 1, (element_indices_j[idx_JI],
                                 element_indices_i[idx_JI]))))
    # NOTE Here, it coincides with Matlab again, though.
    edge_number[idx_JI2IJ - 1] = numbering_IJ - 1

    element_to_edges = edge_number[0:3*n_elements].reshape(n_elements, 3,
                                                           order='F')
    edge_to_nodes = np.column_stack((element_indices_i[idx_IJ],
                                     element_indices_j[idx_IJ]))
    # Provide boundary2edges
    boundaries_to_edges = []
    for j in np.arange(n_boundaries):
        boundaries_to_edges.append(
            edge_number[np.arange(pointer[j+1]+1, pointer[j+2]+1, dtype=int)])
    return element_to_edges, edge_to_nodes, boundaries_to_edges


def get_global_to_local_index_mapping(unique_idxs: np.ndarray):
    """
    given a set of n unique, non-negative integers I,
    this function returns a vectorized and order preserving mapping
    `perform_transform` : I -> [0, 1, 2, ..., n-1] =: I'

    notes
    -----
    - the returned mapping may be used, e.g. to transform global indexation
      to local indexation after extraction of a local patch
    - if global[i, j, ...] <= global[i', j', ...], then we also have
      local[i, j, ...]  <= local[i', j', ...]
    - if N is the number of distinct indices in global_indices,
      then local_indices contains the integers 0, 1, ..., N-1

    example
    -------
    >>> global_indices = np.array([[12, 3, 6],
    >>>                            [12, 2, 6],
    >>>                            [12, 3, 15],
    >>>                            [66, 77, 88]])
    >>> unique_indices = np.unique(global_indices)
    >>> perform_transform = get_global_to_local_index_mapping(unique_indices)
    >>> local_indices = perform_transform(global_indices)
    >>> local_indices
        array([[3, 1, 2],
               [3, 0, 2],
               [3, 1, 4],
               [5, 6, 7]])
    """
    local_idx = np.arange(unique_idxs.size)
    transform = dict(zip(unique_idxs, local_idx))
    perform_transform = np.vectorize(lambda old: transform[old])
    return perform_transform


def complete_boundaries(elements: data_structures.ElementsType,
                        boundaries: list[data_structures.BoundaryType]
                        ) -> list[data_structures.BoundaryType]:
    """
    given a possibly incomplete list of boundaries,
    returns a complete list of boundaries, i.e. adds
    an artificial boundary where needed

    parameters
    ----------
    elements: data_structures.ElementsType
        elements of the mesh at hand
    boundaries: list[data_structures.BoundaryType]
        a possibly incomplete list of boundaries
        of the elements given

    returns
    -------
    completed_boundaries: list[data_structures.BoundaryType]
        a coomplete list of boundaries of the elements given

    example
    -------
    >>> elements = np.array([[0, 1, 2], [0, 2, 3]])
    >>> boundary = np.array([[0, 1], [1, 2]])
    >>> completed_boundaries = complete_boundaries(
    >>>     elements, [boundary])
    >>> completed_boundaries
        [array([[0, 1], [1, 2]]), array([[2, 3], [3, 0]])]
    """
    element_indices_i = elements.flatten()
    element_indices_j = elements[:, [1, 2, 0]].flatten()
    all_edges = np.column_stack([element_indices_i, element_indices_j])

    exterior_edges = []
    for edge in all_edges:
        # mark all rows where edge appears
        edge_is_here = np.sum(np.isin(all_edges, edge), axis=1) == 2
        # if edge appears twice, it is shared by two elements and therefore
        # must be interior
        is_interior = np.sum(edge_is_here) == 2
        if not is_interior:
            exterior_edges.append(edge)

    artificial_boundary = []
    for exterior_edge in exterior_edges:
        covered = False
        for boundary in boundaries:
            if np.any(np.sum(np.isin(boundary, exterior_edge), axis=1) == 2):
                covered = True
                break
        if not covered:
            artificial_boundary.append(exterior_edge)

    if artificial_boundary:
        boundaries.append(np.array(artificial_boundary))
    return boundaries


def get_local_patch(coordinates: data_structures.CoordinatesType,
                    elements: data_structures.ElementsType,
                    boundaries: list[data_structures.BoundaryType],
                    which_for: int
                    ) -> tuple[data_structures.CoordinatesType,
                               data_structures.ElementsType,
                               list[data_structures.BoundaryType]]:
    """
    returns the local mesh corresponding to the k-th element
    and its immediate neightbours, i.e. elements that share an edge
    with the selected k-th element

    parameters
    ----------
    coordinates: data_structures.CoordinatesType
        coordinates of the mesh at hand
    elements: data_structures.ElementsType
        elements of the mesh at hand
    boundaries: list[data_structures.BoundaryType]
        boundaries of the mesh at hand
    which_for: int
        index of the element for which to extract the local patch

    returns
    -------
    local_coordinates: data_structures.CoordinatesType
        coordinates of the selected element and its neighbours,
        where neighbours are elements that share an edge with
        the selected element
    local_elements: data_structures.ElementsType
        neighbours of the selected element, i.e.
        elements that share an edge with
        the selected element
    local_boundaries: list[data_structures.BoundaryType]
        boundaries of the local patch inherited from
        the given boundaries, i.e. if none of the elements
        in the local patch touch the boundary at hand,
        an empty list is returned

    notes
    -----
    the returned local patch is indexed in a local fashion, i.e.
    - the indices in local_elements refer to the entries in local_coordinates
    - the indices in all elements in local_boundaries refer to entries in
      local_coordinates
    """
    # global indices of the marked element's nodes
    nodes = elements[which_for]

    # identifying the marked element's nieghbours
    neighbours = np.sum(np.isin(elements, nodes), axis=1) == 2

    # global indices of the local patch's elements
    local_elements = np.vstack([elements[neighbours], nodes])

    # unique sorted global indices of all nodes in global patch
    unique_idxs = np.unique(local_elements)

    # retreiving the transformation (global -> local indices)
    perform_transform = get_global_to_local_index_mapping(unique_idxs)

    local_boundaries = []
    for boundary in boundaries:
        edge_in_local_patch = np.sum(
            np.isin(boundary, local_elements), axis=1) == 2
        local_boundary = boundary[edge_in_local_patch]
        if local_boundary.size > 0:
            local_boundaries.append(perform_transform(local_boundary))

    # local patch's elements in local indices
    local_elements = perform_transform(local_elements)

    # local patch's coordinates
    local_coordinates = coordinates[unique_idxs]

    return local_coordinates, local_elements, local_boundaries



import numpy as np
from p1afempy import data_structures
from p1afempy import mesh


# TODO refactor: make boundary(conidtions) optional
def compute_eta_r(x: np.ndarray,
                  coordinates: data_structures.CoordinatesType,
                  elements: data_structures.ElementsType,
                  dirichlet: data_structures.BoundaryType,
                  neumann: data_structures.BoundaryType,
                  f: data_structures.BoundaryConditionType,
                  g: data_structures.BoundaryConditionType) -> np.ndarray:
    """
    computes residual-based error estimator for finite element
    solution of Laplace problem with mixed Dirichlet-Neumann
    boundary conditions.
    For details, see chapter 6.2 in the original paper.

    parameters
    ----------
    x: np.ndarray
        the current solution iterate
    coordinates: data_structures.CoordinatesType
        coordinates of the mesh
    elements: elements: data_structures.ElementsType
        elements of the mesh
    dirichlet: data_structures.BoundaryType
    neumann: data_structures.BoundaryType
    f: data_structures.BoundaryConditionType
        function representing Dirichlet BC.
    g: data_structures.BoundaryConditionType
        function representing Neumann BC.

    returns
    -------
    etaR: np.ndarray
        the Mx1 array of squared refinement indicators,
        where M denotes the number of elements.
    """
    boundary_conditions = [dirichlet, neumann]
    element2edges, edge2nodes, boundaries_to_edges = \
        mesh.provide_geometric_data(elements=elements,
                                    boundaries=boundary_conditions)

    # vector of element volumes 2*|T|
    area2 = 2. * mesh.get_area(coordinates=coordinates,
                               elements=elements)

    # compute curl
    d21, d31 = mesh.get_directional_vectors(coordinates=coordinates,
                                            elements=elements)
    tmp1 = x[elements[:, 1]] - x[elements[:, 0]]
    tmp2 = x[elements[:, 2]] - x[elements[:, 0]]
    u21 = np.column_stack([tmp1, tmp1])
    u31 = np.column_stack([tmp2, tmp2])
    curl = (d31 * u21 - d21 * u31) / np.column_stack([area2, area2])

    # compute edge terms hE*(duh/dn) for uh
    dudn21 = np.sum(d21 * curl, axis=1)
    dudn13 = -np.sum(d31 * curl, axis=1)
    dudn32 = -(dudn13 + dudn21)
    etaR = np.bincount(
        element2edges.flatten(order='F'),
        np.hstack([dudn21, dudn32, dudn13]),
        minlength=edge2nodes.shape[0])

    # incorporate Neumann data
    if neumann.size > 0:
        cn1 = coordinates[neumann[:, 0], :]
        cn2 = coordinates[neumann[:, 1], :]
        gmE = g((cn1+cn2) / 2.)
        neumann2edges = boundaries_to_edges[1]
        etaR[neumann2edges] = etaR[neumann2edges] - np.sqrt(np.sum(
            np.square(cn2-cn1),
            axis=1)) * gmE

    # incorporate Dirichlet data
    dirichlet2edges = boundaries_to_edges[0]
    etaR[dirichlet2edges] = 0
    # assemble edge contributions of indicators
    etaR = np.sum(np.square(etaR[element2edges]), axis=1)
    # add volume residual to indicators
    fsT = f((coordinates[elements[:, 0]]+(d21+d31) / 3.))
    etaR = etaR + np.square(0.5 * area2 * fsT)
    return etaR
